/* File: Gauss_model.h
   
   Contains the GMM class called Gauss_model. Fits a GMM to
   the given data. 

   Depenedence: 

   <std_cholesky.h>
*/
#include <iostream>
#include <fstream>
#include <vector>
#include "cholesky.h"
#include <assert.h>
#include <math.h>
#include <limits>
#include "optimization.h"
#include <string>
#include <map>
#include <sstream>
#include <iterator>
#include <algorithm>
#ifndef GAUSS_MODEL_H
#define GAUSS_MODEL_H

template<class T>
inline T SIGN(const T &a, const T &b)
	{return b >= 0 ? (a >= 0 ? a : -a) : (a >= 0 ? -a : a);}

inline float SIGN(const float &a, const double &b)
	{return b >= 0 ? (a >= 0 ? a : -a) : (a >= 0 ? -a : a);}

inline float SIGN(const double &a, const float &b)
	{return (float)(b >= 0 ? (a >= 0 ? a : -a) : (a >= 0 ? -a : a));}

template<class T>
inline T ABS(const T &a)
{return (a >= 0 ?  a : -a);}

inline float ABS(const float &a)
{return (a >= 0 ?  a : -a);}

inline float ABS(const double &a)
{return (float)(a >= 0 ?  a : -a);}


#define SINGLE 1
#define DIAG 2
#define PERT 4

/*
  Constants: Types of Covariance Matrices
  - *MultiFull* Each mixture has its own matrix, which is a full rank matrix. 
  - *SingleFull* Mixtures share a single matrix, which is a full rank matrix.
  - *MultiDiag* Each mixture has its own matrix, which is a diagonal approximation. 
  - *SingleDiag* Mixtures share a single matrix, which is a diagonal approximation. 
  - *MultiPert* Each mixture has its own matrix, which is an low-rank perturbed diagonal decomposition (LRPDD).
  - *SinglePert* Mixtures share a single matrix, which is an low-rank perturbed diagonal decomposition (LRPDD).
 */
#define MultiFull 0
#define SingleFull 1
#define MultiDiag 2
#define SingleDiag 3
#define MultiDiagPert 6
#define SingleDiagPert 7

/*
  Constants: Types of Parameter Initialization
  - 0: *RandomMeans*, the means are randomly choosen from data samples. 
  - 1: *GreeyMeans*, the means are greedily choosen from data samples to be far apart. 
  This approach tends to avoid having two model clusters compete to represent one actual cluster in the data. 
  - 2: *RandomPrunedMeans$*, k log2(k) means are randomly choosen and then pruned down to k means. This
  approach tends to be more robust than random or greedy means. 
  - 3: *LoadMeans*, the means are loaded from a model file. 
 */
#define RandomMeans 0
#define GreedyMeans 1
#define RandomPrunedMeans 2
#define LoadMeans 3

using namespace std;

template<class T>
inline T SQR(const T a) {return a*a;}

template<class T>
inline void SWAP(T &a, T &b)
	{T dum=a; a=b; b=dum;}

// Typedef: DoubMat
// Matrix of type double made of a vector of vectors.  
typedef vector< vector<double> > DoubMat;


void LRP_CG(const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr);

/* 
   Class: Gauss_model

   The GMM class. This class models a given dataset with a mixture of
   multi-dimensional Gaussian distributions.  There are several files
   that are used by Gauss_model and each have their own format.  Below
   the formats are outlined.  

   Format: Data File Format

   - The first line has two values, the number of samples (n) and the sample dimension (m). 
   - The following 'n' lines each correspond to a data sample and contain 'm' values.    

   Format: Model File Format

   - The first line has a single value indicating the number of clusters. 
   - Each cluster is then recorded, as follows (examples in brakets)
   + Cluster id, zero-based, single line [Component 0]
   + Cluster prior probability, single line [Prior 0.3]
   + Mean vector, single line [Mean 1.0 0.5 0.7] 
   + beginning of covariance, single line [Sigma]
   + covariance values, line for each row [0.5 0.2 -0.4]

   Format: Cluster Assignments File Format

   - Each line corresponds to a data sample, in the same order that they occur in the data file
   - Each line has the index of clusters the corresponding data sample was assigned to. 
   
   Format: Prob File Format

   - Each line corresponds to a data sample, in the same order that they occur in the data file
   - Each line has the posterior probabilities of the data sample for each cluster in the model
   
   Format: Sorted Prob File Format

   - Each line corresponds to a data sample/mixture pair, in order of the relative posterior
   probability (defined in <PrintSortedProbs>). 
   - Each line contains the relative probability, data sample name, and mixture id, in that order. 

*/
class Gauss_model {
 public:
  int nn, kk, mm;            // # of data points, components, dimenions
  DoubMat data, means, resp; // x's, mu's, and p(k|n)'s
  vector<double> frac, lndets;      // priori's, log(det(sig_k))
  vector<double> probs,D;
  DoubMat a,diags, beta, olda;
  vector< DoubMat > sig;        // sig_k
  double loglike;              // log(L), log likelihood
  bool isBoosting;
  vector<double> b_wgts;
  vector<double> dists, cum_dist; // for kmeans
  vector<int> count, assign;
  int cov_type;
  int myk;
  double mywgt;

  /*
    Constructor: Gauss_model

    This constructor initializes The GMM model based on the method indicated.

    Parameters:

    ddata - a matrix of the data to model
    style - indicates the type of covariance matrix to use (see <Types of Covariance Matrices>)
    init_type - indicates how to initialize the parameters (see <Types of Parameter Initialization>)
    
   */

  Gauss_model(char* datafilename, int num_clusters, int c_type, int init_type, char* modelfilename = NULL) : 
    kk(num_clusters), cov_type(c_type)
    {      
      // read data
      ReadData(datafilename);

      // prepare parameters
      SetupParameters(init_type);

      if(init_type == 3) { // load modelfile
	LoadModel(modelfilename);
      } else { // initialize model      
	// choose means
	ChooseMeans(init_type);
	
	// if choosen, prune means
	if(init_type == 2) PruneMeans();
      
	// finally, prepare parameters by kmeans
	kmeans();

	// debugging purposes, print out cluster sizes
	ofstream csfile("cluster_size.txt");
	for(int i = 0; i < means.size(); i++) {
	  csfile << count[i] << endl;
	}
	csfile.close();

	// setup posterior probs to reflect partition, and update params
	//for(int n = 0; n < nn; n++)  resp[n][assign[n]] = 1.0;
	//mstep();
      }
    }

  /*
    Function: ReadDmata

    Read in the data from the given filename.

    Parameters: 

    datafilename - char* string with the name of the data file. 

    Preconditions: 

    - datafilename is not null.
    - Data file is of the proper format (see <Data File Format>).     

    Postcondition:
    
    - All data will have been loaded into 'data'.
    - nn, mm are set to the number of samples and sample dimension, respectively. 

  */
  void ReadData(char* datafilename) {
    assert(datafilename != NULL);
   
    ifstream gmmdata(datafilename, ios::in | ios::binary);

    gmmdata >> nn >> mm;
    
    cout << "Grabbing " << nn << " samples ";
    cout << "of dim. " << mm << endl;     

    // Read in data    
    vector<double> tmp(mm,0.0);  
    for(int i = 0; i < nn; i++) {
      for(int j = 0; j < mm; j++) {
	gmmdata >> tmp[j];
      }
      data.push_back(tmp);
    }
    
    gmmdata.close();     
  }


  /*
    Function: ChooseMeans

    Initializes the means of each cluster. The method of initializing depends on the value of 
    init_type (see <Types of Parameter Initialization>). 
   
    Parameters: 

    init_type - integer indicating the type of parameter initialization.

    Preconditions: 

    - init_type is valid. 
    - data has been loaded. 

    Postcondition:

    - For random and greedy initialization, the means have been choosen. For the pruning initialization,
    the k log2(k) means have been choosen. Means are stored in 'means'. 

  */
  void ChooseMeans(int init_type) {
    assert(init_type >= 0 && init_type <= 3); 
    assert(kk > 0); assert(nn > 0);
    assert(mm > 0); assert(data.size() > 0);

    int rand_ind, num_mixes;

    means.clear();

    if( init_type == 0 || init_type == 2) {  // random means
      // set number of means to generate
      num_mixes = kk;
      if(init_type == 2) {
	num_mixes = int(num_mixes * log2(num_mixes)); // k log k 
	num_mixes = (kk > num_mixes) ? kk : num_mixes; 
      }

      // choose means
      for(int i = 0; i < num_mixes; i++) {
	rand_ind = int( ((double)rand()/RAND_MAX) * (double)nn );
	if(rand_ind == nn) rand_ind--;
	means.push_back( data[rand_ind] );    
      }       

    } else if( init_type == 1) { // greedily choose  
      // greedily choose samples as means
      // 1. choose random seed
      rand_ind = int( ((double)rand()/RAND_MAX) * (double)nn );
      if(rand_ind == nn) rand_ind--;
      means.push_back( data[rand_ind] );    
      // 2. setup distance vector
      dists.resize(nn,numeric_limits<double>::infinity());
      // 3. for each cluster, find farthest sample
      double sdist, max_dist;
      int max_ind; 
      for(int k = 0; k < kk - 1; k++) { 
	max_dist = 0;
	for(int i = 0; i < nn; i++) {
	  sdist = 0;
	  // calc distance
	  for(int m = 0; m < mm; m++) { sdist += (data[i][m] - means[k][m])*(data[i][m] - means[k][m]); }
	  // if distance is less than previous, record it
	  dists[i] = (dists[i] < sdist) ? dists[i] : sdist;
	  // keep running track of maximum distance sample
	  if(dists[i] > max_dist) { 
	    max_dist = dists[i];
	    max_ind = i;
	  }
	}
	// set next sample as farthest sample
	means.push_back(data[max_ind]);
	}
      dists.clear(); // clear for use by kmeans
     
    }
  }

  /*
    Function: PruneMeans

    Called when performing intializing the means by pruning from k log(k) means. 
    K-means clustering is performed with the k log(k) means. Clusters of a size 
    below 0.5*(nn/(k log k)) + 2 are discarded. If more than k means remain, 
    the function chooses k means that are farthest apart from each other. Otherwise,
    the remaining means are used as the inital means. Note: there may be fewer than
    k means, if the initial k-mean clusters are mostly sufficiently small. 

    Parameters:

    none

    Preconditions: 
    
    - k log(k) means have been choosen and stored in 'means'. 

    Postcondition:

    - k or fewer means remain in 'means'.
    - 'kk' has been updated, if fewer than k means remained. 

  */
  void PruneMeans() {
    assert(kk == 1 || means.size() == int(kk * log2(kk)));

    // optimize them
    kmeans();    
    cout << "Initialized " << means.size() << " clusters" << endl;

    // 2a. remove 'small' clusters, where small is less than 0.5*(nn/(k log k)) + 2
    int small_size = int(0.5*nn/(kk*log2(kk)) + 2);
    cout << "Thresholding clusters to have at least " << small_size << " nodes" << endl;
    for(int i = means.size() - 1; i >= 0 ; i--) {
      if( count[i] < small_size) {
	means.erase( means.begin() + i);
      }
    }
    int num_means = means.size(); // update value
    if(num_means < kk) { kk = num_means; } 
    cout << "Working with " << kk << " clusters" << endl;


    // prune down to kk clusters
    // greedily choose samples as means
    // 1. choose random seed    
    int rand_ind = int( ((double)rand()/RAND_MAX) * (double)num_means );
    if(rand_ind == num_means) rand_ind--;
    means.push_back( means[rand_ind] );    
    // 2. setup distance vector    
    vector<double> mdists(num_means,numeric_limits<double>::infinity());    

    // 3. for each cluster, find farthest sample    
    double sdist, max_dist;
    int max_ind; 
    for(int k = 0; k < kk - 1; k++) {      
      max_dist = 0;
      for(int i = 0; i < num_means; i++) {
	sdist = 0;
	// calc distance
	for(int m = 0; m < mm; m++) { sdist += (means[i][m] - means[num_means+k][m])*(means[i][m] - means[num_means+k][m]); }
	// if distance is less than previous, record it
	mdists[i] = (mdists[i] < sdist) ? mdists[i] : sdist;
	// keep running track of maximum distance sample
	if(mdists[i] > max_dist) { 
	  max_dist = mdists[i];
	  max_ind = i;
	}
      }
      // set next sample as farthest sample
      means.push_back(means[max_ind]);
    }

    cout << "removing extra centers" << endl;
    while(means.size() > kk ) {
      means.erase( means.begin() );
    } 
  }
  
  /*
    Function: SetupParameters

    Allocates memory for all variables needed to train the model. 

    Parameters: 

    init_type - integer indicating the type of parameter initialization.

    Preconditions: 

    - size and dimension of the data set has been obtained.
    - number of clusters has been obtained. 

    Postcondition:

    - variables used for training have allocated memory space. 
    - Covariances are set to I*1e-5, where I is the identity matrix. 

  */
  void SetupParameters(int init_type) {
    // setup space
    resp.resize(nn);
    frac.resize(kk); 
    lndets.resize(kk);
    sig.resize(kk);
    probs.resize(nn);
    a.resize(kk);
    D.resize(mm);
    diags.resize(kk);
    olda.resize(kk);
    beta.resize(nn);
    dists.resize(nn);
    cum_dist.resize(nn);
    count.resize(kk);
    assign.resize(nn);

    isBoosting = false;

    int i,j,k;

    // setup resp (nn,kk), sig(kk,mm,mm), a(kk,mm)
    vector<double> tmp(kk,0);
    for(i = 0; i < nn; i++) {
      resp[i] = tmp;
      beta[i] = tmp;
    }
    tmp.resize(mm,0);
    for(i = 0; i < kk; i++)
      {
	a[i] = tmp;
	diags[i] = tmp;
	sig[i].resize(mm,tmp);
	olda[i] = tmp;
      }
    

    // for each component
    std::cout << "Initializing GMM parameters" << endl;
    for( k = 0; k < kk; k++) {
      frac[k] = 1./kk;       // uniform priors
      for( i = 0; i < mm; i++) { // low-variance, indep. sigma , \Sigma = I * e
	// prepare diagonals for perturbed matrices
	diags[k][i] = 1.0;
	for(j = 0; j < mm; j++) sig[k][i][j] = 0.;
	sig[k][i][i] = 1.0e-5;  // not changed
      }      
    }    
    for(i = 0; i < mm; i++) { D[i] = 0; } // initialize D


    cout << "Training GMM with " << kk << " " << mm << "-dim. components using ";
    cout << nn << " samples" << endl;

  }

  /*
    Function: LoadModel

    Loads a set of model parameters from a given file. 

    Parameters: 

    modelfilename - a string of the name of the model file. 

    Preconditions: 
    
    - modelfilename is not null. 
    - model file is in the correct format (see <Model File Format>). 

    Postcondition:

    - the prior probabilities, means and covariances for each cluster 
    are loaded into the model. 

  */
  void LoadModel(char* modelfilename) { 
    assert(modelfilename != NULL);
    isBoosting = false;

    //std::cout << "Constructing gmm model (testing)" << endl;
    ifstream gmmfile(modelfilename);
    gmmfile.setf(ios::scientific,ios::floatfield);
    gmmfile.precision(30);
    int it;
    double dt;
    std::string st;    
          
    // get model info
    gmmfile >> it;
    //kk = it; 
    vector<double> tmp(mm,0.0);    

    means.resize(kk,tmp);    
    tmp.resize(mm,0.0);
    a.resize(kk,tmp);
    frac.resize(kk);
    sig.resize(kk);
    tmp.resize(kk,0.0);
    resp.resize(nn,tmp);
    lndets.resize(kk);
    tmp.resize(mm,0.0);
    
    
    for(int i = 0; i < kk-1; i++) {
      //std::cout << "reading in component" << i;      
      sig[i].resize(mm,tmp);
      gmmfile >> st >> it; // read in 'Component k' where k should be = i
      //std::cout << "\tprior";
      gmmfile >> st >> dt; frac[i] = dt; // get prior
      //std::cout << "\tmean";
      gmmfile >> st; // read in mean vector
      for(int m = 0; m < mm; m++) {
	gmmfile >> dt; means[i][m] = dt;
      }
      //std::cout << "\tsig";
      gmmfile >> st; // read in sig
      for(int m = 0; m < mm; m++) {
	for(int mp = 0; mp < mm; mp++) {
	  gmmfile >> dt; sig[i][m][mp] = dt;  
	}         
      }
      //std::cout << endl;
    }

    gmmfile.close();
  
    // reset priors
    for(int k = 0; k < kk; k++) frac[k] = 1.0/kk;

    // choose new mean
    int rand_ind = int( ((double)rand()/RAND_MAX) * (double)nn );
    if(rand_ind == nn) rand_ind--;
    //means.push_back( data[rand_ind] );
    for(int m = 0; m < mm; m++) means[kk-1][m] = data[rand_ind][m];

    // add new cov. matrix
    sig[kk-1].resize(mm,tmp);
    for( int i = 0; i < mm; i++) { // low-variance, indep. sigma , \Sigma = I * e
      for(int j = 0; j < mm; j++) sig[kk-1][i][j] = 0.;
      sig[kk-1][i][i] = 1.0e-5;
    } 

    cout << "Opened GMM with: " << kk << " " << mm << "-dim. components" << endl;
    cout << "\t\t" << nn << " data samples " << endl;
  }


  /*
    Function: kmeans

    Performs k-means clustering on the data. This is used for initializing the means and covariance matrices. 
    The function stores the resulting means in the 'means' variable and the assignments in the 'assign' vector. 

    Parameters:
    
    none

    Precondition:

    - data has been load.
    - means have been chosen. 

    Postcondition:

    - means have been improved w.r.t. within-cluster sum of squares. 
    - cluster sizes have been stored in 'count'. 

    See also:
    
    <kmeans_dists>, <kmeans_e>, <kmeans_m>
   */
  void kmeans() {
    // initialize distances
    count.clear();
    count = vector<int>(means.size(),0);
    cout << "init kmeans" << endl;
    //cout << "nn=" << nn << endl ;
    for(int n = 0; n < nn; n++) {
      dists[n] = 9.99e99; // set all distances to max
      assign[n] = n%kk; // evenly distribute points amongst means
    }

    // perform kmeans
    cout << "performing kmeans" << endl;
    for(int i = 0; i < 15; i++) {  // why 15 ?
      cout << "Moved " << kmeans_e() << " data samples" << endl;
      kmeans_m();
    }
  }

  /*
    Function: kmeans_dists

    Calculates distances from data samples to means.

    Parameters: 

    k - index of cluster to compute distances for. 

    Precondition:

    - k is a valid index
    - distances from cluster k's mean to all nodes has been computed 
    and stored in cum_dist. 
    
   */
  void kmeans_dists(int k) { // calculate distances to means[k]    
    int m,n;
    double d;    

    for (n=0;n<nn;n++) {  
	for (d=0.,m=0; m<mm; m++) d += SQR(data[n][m]-means[k][m]);
	if (d < dists[n]) {dists[n] = d;}	
	if(n == 0 )
	  cum_dist[n] = dists[n];
	else
	  cum_dist[n] = cum_dist[n-1] + dists[n];
    }
  }  

  /*
    Function: kmeans_e 

    Expectation step of k-means clustering. Clusters data using given means. 

    Parameters: 

    none

    Preconditions:

    - data and means have been loaded/chosen. 

    Postconditions:

    - all data samples have been assigned to a cluster, which
    is recorded in 'assign'. 

   */
  int kmeans_e() {
    int k,m,n,kmin;
    double dmin,d;
    int nchg = 0;
    for (k=0;k<means.size();k++) count[k] = 0;
    for (n=0;n<nn;n++) {
      // for each sample, find closest mean
      for (k=0;k<means.size();k++) {       
	for (d=0.,m=0; m<mm; m++) { d += SQR(data[n][m]-means[k][m]) ; }
	// cout << "d=" << d << " ";
	if(k == 0) {dmin = d; kmin = k;}
	if (d < dmin) {dmin = d; kmin = k;}
      }
      //cout << endl ;	
      // assign sample to appropriate mean
      if (kmin != assign[n]) nchg++;
      assign[n] = kmin;
      count[kmin]++;
    }
    return nchg;
  }

  /*
    Function: kmeans_m
    
    Maximization step of k-means clustering. Determines means based on data assigned to cluster. 

    Parameters:

    none

    Preconditions:

    - data has been loaded. means have been chosen. 
    - data samples have been assigned to clusters

    Postconditions:

    - cluster means have been updated based on the data samples 
    within the cluster. 

  */
  void kmeans_m() {
    int n,k,m;
    for (k=0;k<means.size();k++) for (m=0;m<mm;m++) means[k][m] = 0.; // clear means
    // set means as average of all associated points
    for (n=0;n<nn;n++) for (m=0;m<mm;m++) means[assign[n]][m] += data[n][m]; 
    for (k=0;k<means.size();k++) {
      if (count[k] > 0) for (m=0;m<mm;m++) means[k][m] /= count[k]; 
    }
  }

  /*
    Function: SetBoostWeights

    This function sets the Boosting weights to be set/updated. These weights
    allow for additional weighting, as used in Adaboost. 

    Parameters:

    weights - vector of length 'nn' that holds weights

    Preconditions:

    - weights contains a weight for each data sample

    Postcondition:

    - weights are loaded. 

   */
  void SetBoostWeights(vector<double> &weights) {
    // cout << "GMM boost weights set" << endl;
    isBoosting = true;
    b_wgts.clear();
    
    // cout << "resize to " <<  (int)weights.size() << endl;
    /*
    try {
      b_wgts.resize((int)weights.size(),0.0);
    } catch (bad_alloc const&) {
      cout << "Memory allocation fail!" << endl;
      exit(0);
    }    
    
    cout << "copy" << endl;
    copy(weights.begin(),weights.end(),b_wgts.begin());
    */

    b_wgts.reserve((int)weights.size());
    // cout << "copy" << endl;
    for(int i = 0; i < (int)weights.size(); i++) {
      b_wgts.push_back(weights[i]);
    }
    // cout << "done" << endl;
  }

  /*
    Function: estep

    Expectation step of GMM training. This function calculates the posterior probabilities 
    of each data sample for each mixture in the model, given the model parameters. 

    Parameters: 

    none

    Preconditions:

    - data has been loaded.
    - model parameters have been loaded or initialized.

    Postconditions:

    - posterior probabilites for each data sample and cluster has been computed. 
    - loglikelihood of model with current parameters is computed. 

   */
  double estep() { // one E-step
    //cout << "E-step:" << endl;
    int k,m,n;
    double tmp,sum,max,oldloglike;    
    vector<double> u(mm,0);
    vector<double> v(mm,0);
    double p,b;

    oldloglike = loglike; // hold old loglike
    cout << "Cov_type: " << cov_type << endl;

    // Finding logs of pnks
    for(k = 0; k < kk; k++) { // for each component
      if( (cov_type == 6) || (cov_type == 7) ) {
	// find log det		
	if( (cov_type == 7) && k != 0) {
	  lndets[k] = lndets[0];
	} else {
	  p = 0; b = 0;
	  for(m = 0; m < mm; m++) p += 2.0*log(SQR(diags[k][m]));    
	  for(m = 0; m < mm; m++) b += exp(p - 2.0*log(SQR(diags[k][m])) + (a[k][m] < 1.0e-6 ? 0 : log(SQR(a[k][m])))  ); // e-6 changed to e-10 by SP
	  lndets[k] = -log(b+ exp(p)); // the negative is because D^2 + a*a' is the inverse of Sigma
	}
	
	for(n = 0; n < nn; n++) {
	  // find (x(n)-mu(k))'*(D^2+a*a')*(x(n)-mu(k))
	  if( (cov_type == 7) ) {
	    for(sum = 0.,m=0; m < mm; m++) sum += (data[n][m]-means[k][m])*(a[0][m]);
	    sum = SQR(sum);
	    for(m= 0; m < mm; m++) sum += SQR(data[n][m]-means[k][m])*SQR(diags[0][m]);
	  } else {
	    for(sum = 0.,m=0; m < mm; m++) sum += (data[n][m]-means[k][m])*(a[k][m]);
	    sum = SQR(sum);
	    for(m= 0; m < mm; m++) sum += SQR(data[n][m]-means[k][m])*SQR(diags[k][m]);
	  }

	  if(isBoosting) {
	    //cout << "Boosting" << endl;
	    resp[n][k] = log(b_wgts[n])-0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	  } else {
	    resp[n][k] = -0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	  }	
	}
      } else if( (cov_type >= 2) && (cov_type <= 3) ) { // diagonal matrices
	// find log det, log of the product of the diagonal elements squared
	lndets[k] = 0;
	for(m = 0; m < mm; m++) lndets[k] += log(sig[k][m][m]);
	lndets[k] *= 2.0; 

	for(n = 0; n < nn; n++) {
	  // find (x(n)-mu(k))'*(D^-1)*(x(n)-mu(k))
	  for(sum = 0.,m=0; m < mm; m++) sum += SQR(data[n][m]-means[k][m])/sig[k][m][m];
	  if(isBoosting) {
	    //cout << "Boosting" << endl;
	    resp[n][k] = log(b_wgts[n])-0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	  } else {
	    resp[n][k] = -0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	  }
	}
      } else { // full matrix
	//cout << "\t\tComponent " << k << endl;
	try{ 
	  Cholesky choltmp(sig[k]); // setup for cholesky decomp. on sigma
	  lndets[k] = choltmp.logdet(); // get log. determinant of sigma	  
	  for (n=0;n<nn;n++) { // Inner loop for pnk's.
	    for (m=0;m<mm;m++) u[m] = data[n][m]-means[k][m]; // find distance from mean
	  
	    choltmp.Lsolve(u,v); // Solve L . v = u.		    	    
	    for (sum=0.,m=0; m<mm; m++) sum += v[m]*u[m]; // solve v = u . L^-1 . u^T  

	    if(isBoosting) {
	      //cout << "Boosting" << endl;
	      resp[n][k] = log(b_wgts[n])-0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	    } else {
	      resp[n][k] = -0.5*(sum + lndets[k]) + log(frac[k]); // Eq. 16.1.5, but unnormalized
	    }	
	    //if(k == 0) { probs[n] = resp[n][k]; } { probs[n] += resp[n][k]; }
	  }
	} catch (char const* str) {
	  cout << "Caught exception in estep" << endl;
	  cout << "Cholesky exception: " << str << endl;
	  cout << "Fraction: " << frac[k] << endl;
	  	  
	ofstream wfile("wfile.txt");
	wfile << "problem in component k = " << k << endl;
	for(n = 0; n < nn; n++) {
	  if(resp[n][k] > 0) {
	    wfile << "probs: ";
	    for(int j=0;j<kk;j++) 
	      wfile << resp[n][j] << " ";
	    wfile << endl << "data: ";
	    for(m = 0; m < mm; m++) {
	      wfile << data[n][m] << " ";
	    }
	    wfile << endl;
	  }
	}
	wfile.close();
	  exit(0);
	}      
      }
    }

    // At this point we have unnormalized logs of the pnk's. We need to normalize using
    // log-sum-exp and compute the log-likelihood.
    //cout << "\tNormalizing logs of pnks" << endl;
    loglike = 0;
    for (n=0;n<nn;n++) { // Separate normalization for each n.
      max = -99.9e99; // Log-sum-exp trick begins here.
      for (k=0;k<kk;k++) if (resp[n][k] > max) max = resp[n][k]; // find z_max
      for (sum=0.,k=0; k<kk; k++) sum += exp(resp[n][k]-max); // second right hand term in 16.1.9
      tmp = max + log(sum); // Eq 16.1.9

      probs[n] = (double)tmp;
      for (k=0;k<kk;k++) {
	resp[n][k] = exp(resp[n][k] - tmp); // normalize pnk
	//if(resp[n][k] < std::numeric_limits<double>::epsilon()) {
	  //cout << "found near zero post. prob: (n k) = " << n << k << endl;
	  //resp[n][k] = 0;
	//}	  
      }
      loglike +=tmp; // add log likelihood of sample n
    }
    cout << "Change in Log like.: " <<  (loglike-oldloglike) << endl;
    return loglike - oldloglike; // When abs of this is small, then we have    
  } //converged.


  /* 
     Function: mstep_MultiFull

     This function handles the maximization step when the model uses unique and full rank
     covariance matrices for each mixture. Pre- and Post-conditions are the same as 
     for <mstep>. 

     Parameters:

     none    

   */
  void mstep_MultiFull() {
    // Perform one M-step of the EM algorithm.
    int j,n,k,m;
    double wgt,sum;
    cout << "mstep:" << endl;

    // clear D
    for(m=0;m<mm;m++) { D[m] = 0.0;}

    for (k=0;k<kk;k++) { // for each componenet
      //cout << "\tWorking on component " << k << endl;
      wgt=0.;
      for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
      frac[k] = wgt/nn; // Equation (16.1.7).
      //cout << "fraction: " << frac[k] << endl;
      for (m=0;m<mm;m++) { // for each dimension
	// update mean
	for (sum=0.,n=0; n<nn; n++) {
	  D[m] += resp[n][k]*SQR(data[n][m]-means[k][m]);
	  sum += resp[n][k]*data[n][m]/wgt; // weighted sum of data
	}
	means[k][m] = sum; // Equation (16.1.6a)., sum normalized
	// update cov.
	for (j=0;j<=m;j++) { 
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*(data[n][m]-means[k][m])*(data[n][j]-means[k][j])/wgt; // part of Eq. 16.1.6b
	  }
	  //sig[k][m][j] = sum/wgt; // Equation (16.1.6b.)
	  //sig[k][j][m] = sum/wgt; // Equation (16.1.6b).
	  sig[k][j][m] = sig[k][m][j] = sum; // Equation (16.1.6b.)

	}
      }

      // Cholesky Decomp
      try{ 
	Cholesky choltmp(sig[k]); 
      } catch (char const* str) {
	cout << "Caught exception in mstep" << endl;
	cout << "Cholesky exception: " << str << endl;
	cout << "Fraction: " << frac[k] << endl;
	cout << "Mean: " << endl;
	for(m = 0; m < mm; m++) {
	//  cout << means[k][m] << " ";
	}
	cout << endl;

	for(m = 0; m < mm; m++) {
	  for(int mp = 0; mp < mm; mp++) {
	  //  cout << sig[k][m][mp] << " ";  
	  }         
	  cout << endl;
	}
	ofstream wfile("wfile.txt");
	wfile << "problem in component k = " << k << endl;
	for(n = 0; n < nn; n++) {
	  if(resp[n][k] > 0) {
	    wfile << "probs: ";
	    for(j=0;j<kk;j++) 
	      wfile << resp[n][j] << " ";
	    wfile << endl << "data: ";
	    for(m = 0; m < mm; m++) {
	      wfile << data[n][m] << " ";
	    }
	    wfile << endl;
	  }
	}
	wfile.close();
	exit(0);
      }     
    }
  }


  /*
    Function: mstep

    Maximization step for GMM training. This function handles updating the model's parameters
    based on posterior probabilities of the data for each mixture. Since there are mulitple 
    methods for updating the covariance matrix, this function is essentially a switch statement. 

    Parameters: 

    none
    
    Preconditions:

    - data, model parameters are loaded and valid. 
    - posterior probabilities have been computed for this parameter set. 
     
    Postconditions:

    - model parameters (prior probs, means, covariances) are updated 
    so as to maximize the model likelihood. 

    See also:
    
    <mstep_MultiFull> 
   */
  void mstep(double scale_parameter) {
    // Perform one M-step of the EM algorithm.
    cout << "Now in M-step:" << endl;
    int j,n,k,m,m2;
    double wgt,sum,total_wgt;

    // multi-perturbed
    double muTa;
    double W;
    //vector<double> beta;
    double C,anorm, diffnorm;
    //vector<double> olda;
    int iters =0;
    double matdiff, denom;

    vector<double> ga(mm,0);
    vector<double> gd(mm,0);
    vector<double> ga1(mm,0);
    vector<double> gd1(mm,0);

    vector<double> da(mm,0);
    vector<double> dd(mm,0);

    double betat,beta_denom;
    double gnorm, step;


    // end

    // clear D
    for(m=0;m<mm;m++) { D[m] = 0.0;}
    total_wgt = 0.;
    cout<< cov_type << endl;
    switch(cov_type) {
    case MultiFull:
      cout << "multifull" << endl;
      mstep_MultiFull();
      break;

    case SingleDiag:
      // determine diagonal cov

      // copy over to other covs
      for (k=0;k<kk;k++) { // for each componenet
	//cout << "\tWorking on component " << k << endl;
	wgt=0.;
	for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
	total_wgt += wgt;
	frac[k] = wgt/nn; // Equation (16.1.7).
	for (m=0;m<mm;m++) { // for each dimension	
	  // update mean
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*data[n][m]; // weighted sum of data
	  }
	  means[k][m] = sum/wgt; // Equation (16.1.6a)., sum normalized

	  // update covariance
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*
	      (data[n][m]-means[k][m])*(data[n][m]-means[k][m]); // part of Eq. 16.1.6b
	  }
	  
	  for(int kit = 0; kit < kk; kit++) {
	    if(k == 0)
	      sig[kit][m][m] = frac[k]*sum/wgt; // Equation (16.1.6b.)	  
	    else
	      sig[kit][m][m] += frac[k]*sum/wgt;
	  }
	}
      }
      break;

    case SingleFull:

      for (k=0;k<kk;k++) { // for each componenet
	//cout << "\tWorking on component " << k << endl;
	wgt=0.;
	for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
	total_wgt += wgt;
	frac[k] = wgt/nn; // Equation (16.1.7).
	for (m=0;m<mm;m++) { // for each dimension	
	  // update mean
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*data[n][m]; // weighted sum of data
	  }
	  means[k][m] = sum/wgt; // Equation (16.1.6a)., sum normalized

	  // update covariance
	  for(j = 0; j <= m; j++) {
	    // 1) sum over cov for all covariances
	    for (sum=0.,n=0; n<nn; n++) {
	      sum += resp[n][k]*
		(data[n][m]-means[k][m])*(data[n][j]-means[k][j]); // part of Eq. 16.1.6b
	    }	  
	    // 2) normalize sum and set covariances
	    for(int kit = 0; kit < kk; kit++) {	      
	      if(k == 0) {
		sig[kit][m][j] = frac[k]*sum/wgt; // Equation (16.1.6b.)	  
		sig[kit][j][m] = frac[k]*sum/wgt; // Equation (16.1.6b.)	  
	      } else {
		sig[kit][m][j] += frac[k]*sum/wgt;
		if(j != m) // don't update diagonals twice
		  sig[kit][j][m] += frac[k]*sum/wgt;
	      }
	    }
	  }
	}
      }
      break;

    case MultiDiag:
      for (k=0;k<kk;k++) { // for each componenet
	//cout << "\tWorking on component " << k << endl;
	wgt=0.;
	for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
	total_wgt += wgt;
	frac[k] = wgt/nn; // Equation (16.1.7).
	for (m=0;m<mm;m++) { // for each dimension	

	  // update mean
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*data[n][m]; // weighted sum of data
	  }
	  means[k][m] = sum/wgt; // Equation (16.1.6a)., sum normalized

	  // update covariance
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*
	      (data[n][m]-means[k][m])*(data[n][m]-means[k][m]); // part of Eq. 16.1.6b
	  }
	  sig[k][m][m] = sum/wgt; // Equation (16.1.6b.)	
	}
      }
      break;
      
    case MultiDiagPert:     
      // approximation of covariances
      for(k = 0; k < kk; k++) {	
	//cout << "\tWorking on component " << k << endl;
	wgt=0.;
	for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
	total_wgt += wgt;
	frac[k] = wgt/nn; // Equation (16.1.7).
	for (m=0;m<mm;m++) { // for each dimension	

	  // update mean
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*data[n][m]; // weighted sum of data
	  }
	  means[k][m] = sum/wgt; // Equation (16.1.6a)., sum normalized

	  // update covariance
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*
	      (data[n][m]-means[k][m])*(data[n][m]-means[k][m]); // part of Eq. 16.1.6b
	  }
	  sig[k][m][m] = sum/wgt; // Equation (16.1.6b.)	
	}

	double trs = 0 ;
	//cout <<" printing sig" << endl ;
        for (m=0;m<mm;m++) {
                 trs = trs + sig[k][m][m] ;
        }
	trs = (scale_parameter*trs)/mm ;
	// Initialize

	
	// the low rank perturbation and diagonal matrix
	// are stored as a single array for conj. grad. 
	alglib::real_1d_array lr_pert;
	
	lr_pert.setlength(2*mm);
	for(m = 0; m < mm; m++) {	
	  // generate random vector for perturbation
	  lr_pert[m] = (double)rand()/RAND_MAX;	  
	  // setup diagonal [ diag(Sigma) ]
	  // and store diagonals in sig, for gradient calcs
	  if (sig[k][m][m]!=0) {  // change made by SP
	  	lr_pert[m+mm] = sqrt(1.0/sig[k][m][m]); // not changed by sp
	  } else {  
		lr_pert[m+mm] = trs ;  //change made by SP
	  }

	}

	
	double epsg = 0.0001;
	double epsf = 0;
	double epsx = 0;
	
	myk = k;
	mywgt = wgt;
	
	alglib::ae_int_t maxits = 10;
	alglib::mincgstate state;
	alglib::mincgreport rep;
	//cout << "Step 0" ;
	alglib::mincgcreate(lr_pert, state);
	//cout <<" Step 1" ;
	alglib::mincgsetcond(state, epsg, epsf, epsx, maxits);
	//cout <<"Step 2";	
	alglib::mincgoptimize(state, LRP_CG, NULL, (void *)this);
	//cout << "Step 3" << endl ;	
	alglib::mincgresults(state, lr_pert, rep);

	cout << "CG return: " << int(rep.terminationtype) << endl; 

	// copy result from alglib type to vector
	for(m = 0; m < mm; m++) {
	  a[k][m] = lr_pert[m];
	  diags[k][m] = lr_pert[m+mm];
	}
	
	cout << "update iterations: " << iters << endl;

	// update covariance	
	denom = 1;
	for(m = 0; m < mm; m++) {
	  denom += a[k][m]*a[k][m]/(diags[k][m]*diags[k][m]);
	}

	cout << "Updating covariance" << endl;
	for(m = 0; m < mm; m++) {
	  for(m2 = 0; m2 <= m; m2++) {	    
	    sig[k][m][m2] = -a[k][m]*a[k][m2]/(SQR(diags[k][m]*diags[k][m2]))/denom;
	    if(m2 == m) sig[k][m][m2] += 1.0/SQR(diags[k][m]);
	    else sig[k][m2][m] = sig[k][m][m2];
	  }
	}	
      }
      cout << "done with components" << endl;
      break;

    
    case SingleDiagPert:
      // approximation of covariances
      for(k = 0; k < kk; k++) {	
	cout << "\tWorking on component " << k << endl;
	wgt=0.;
	for (n=0;n<nn;n++) wgt += resp[n][k]; // wgt = P(x_n)
	total_wgt += wgt;
	frac[k] = wgt/nn; // Equation (16.1.7).
	for (m=0;m<mm;m++) { // for each dimension	

	  // update mean
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*data[n][m]; // weighted sum of data
	  }
	  means[k][m] = sum/wgt; // Equation (16.1.6a)., sum normalized
	 // cout <<"sum="<< sum <<"wgt=" <<wgt << endl ;
	  // update covariance
	  for (sum=0.,n=0; n<nn; n++) {
	    sum += resp[n][k]*
	      (data[n][m]-means[k][m])*(data[n][m]-means[k][m]); // part of Eq. 16.1.6b
	  }
	  sig[k][m][m] = sum/wgt; // Equation (16.1.6b.)	
	}
	double trs = 0 ;
	//cout <<" printing sig" << endl ;
        for (m=0;m<mm;m++) {
                 trs = trs + sig[k][m][m] ;
        }
	trs = (scale_parameter*trs)/mm ;

	// the low rank perturbation and diagonal matrix
	// are stored as a single array for conj. grad. 
	alglib::real_1d_array lr_pert;
	
	lr_pert.setlength(2*mm);
	for(m = 0; m < mm; m++) {	
	  // generate random vector for perturbation
	  lr_pert[m] = (double)rand()/RAND_MAX;	  
	  // setup diagonal [ diag(Sigma) ]
	  // and store diagonals in sig, for gradient calcs
	  if (sig[k][m][m]!=0) {
		  lr_pert[m+mm] = sqrt(1.0/sig[k][m][m]); // changed by sp/
	  } else {
		 lr_pert[m+mm] = trs;
	  }
	//  cout << lr_pert[m] << " " << lr_pert[m+mm] <<" " ;
	}
	//cout << endl ;
	double epsg = 0.0001;
	double epsf = 0;
	double epsx = 0;
	
	myk = k;
	mywgt = wgt;
	
	alglib::ae_int_t maxits = 5;
	alglib::mincgstate state;
	alglib::mincgreport rep;

	alglib::mincgcreate(lr_pert, state);
	alglib::mincgsetcond(state, epsg, epsf, epsx, maxits);
	alglib::mincgoptimize(state, LRP_CG, NULL, (void *)this);
	alglib::mincgresults(state, lr_pert, rep);

	cout << "CG return: " << int(rep.terminationtype) << endl; 

	// copy result from alglib type to vector
	for(m = 0; m < mm; m++) {
	  a[k][m] = lr_pert[m];
	  diags[k][m] = lr_pert[m+mm];
	}


	// update covariance
	/*
	  denom = 1;
	  for(m = 0; m < mm; m++) {
	  denom += a[k][m]*a[k][m]/(diags[k][m]*diags[k][m]);
	  }

	  cout << "Updating covariance" << endl;
	  for(m = 0; m < mm; m++) {
	  for(m2 = 0; m2 <= m; m2++) {	    
	  sig[k][m][m2] = -a[k][m]*a[k][m2]/(SQR(diags[k][m]*diags[k][m2]))/denom;
	  if(m2 == m) sig[k][m][m2] += 1.0/SQR(diags[k][m]);
	  else sig[k][m2][m] = sig[k][m][m2];
	  }
	  }
	
	  }*/
      }
    	
      // to combine them into one, weighted sum across a and D
      // final answer is stored in the first component
      for(k = 0; k < kk; k++) {
	for(m = 0; m < mm; m++) {
	  if(k == 0) { 
	    a[k][m] *= frac[k]/wgt;
	    diags[k][m] *= frac[k]/wgt;
	  } else {
	    a[0][m] += frac[k]*a[k][m]/wgt;
	    diags[0][m] += frac[k]*diags[k][m]/wgt;
	  }	      
	}
      }

      //cout << "done with components" << endl;
      break;     
    }	
    //cout << "done with m" << endl;
  }

  

  /*
    Function: SigmaMult

    Calculates y = Sigma * x in linear time. This is used with low-rank perturbed diagonal decomposition (LRPDD). 

    Parameters:

    x - input vector (size mm)
    y - holds result of multiplication (size mm)
    k - indicates the mixture that Sigma is taken from
    wgt - the prior probability of this mixture

    Preconditions:

    - input vectors are of appropriate size
    - data, means, posterior probs have been loaded/computed.     

    Postconditions:

    - y contains the result of Sigma * x
    
   */
  void SigmaMult(vector<double> &x, vector<double> &y, int k, double wgt) {
    double muTa = 0;
    int n,m;

    for(m=0; m < mm; m++) {
      // project onto mean
      muTa += x[m] * means[k][m];  
      y[m] = 0;
    }

    double W = wgt;
    double C = 0;

    for(n = 0; n < nn; n++) {
      // set beta values
      beta[n][k] = 0;
      for(m = 0; m < mm; m++) {
	beta[n][k] += data[n][m]*x[m];
      }
      beta[n][k] = resp[n][k]*(beta[n][k]-muTa);

      // sum weighted samples	    
      for(m = 0; m < mm; m++) {
	//y[m] += beta[n][k]*data[n][m]/W;
	y[m] += beta[n][k]*data[n][m];
      }
      // sum beta values
      C += beta[n][k];
    }
	  
    // update vector
    for(m = 0; m < mm; m++) {
      //y[m] = y[m] - means[k][m]*C;
      y[m] = (y[m] - means[k][m]*C)/W;
    }
  }
 
  /* 
     Function: GetGradient

     Calculates the gradient in parameter space of the current parameter values. 
     This is used for the conjugate gradients technique in training the
     LRPDD parameters. 

     Parameters:

     ga - the gradient for the low-rank perturbation (size dd)
     gd - the gradient for the diagonal matrix (size dd)
     k - the mixture for which we are calculating the gradient 
     wgt - the prior probability of this mixture

     Preconditions:

     - input vectors are of appropriate size
     - diagonal matrix, 'diags',  and matrix perturbation, 'a', have been computed. 

     Postconditions:
     
     - 'ga' and 'gd' hold the gradient of the parameters of mixture 'k'. 
   */
  void GetGradient(vector<double> &ga, vector<double> &gd, int k, double wgt) {
    int n,m;
    double denom = 1;
    double aDa = 0;
    vector<double> tmp(mm,0);
    if(ga.size() != mm) ga.resize(mm,0.0);
    if(gd.size() != mm) gd.resize(mm,0.0);

    // calculate denominator
    for(m = 0; m < mm; m++) {
      tmp[m] = SQR(a[k][m])/SQR(diags[k][m]);
      aDa += tmp[m];
    }
    denom = 1 + aDa;

    // find alpha's gradient
    // ga = 2*Sigma*a - 2*D^-2*a + 2*D^-2*a*a'*D^-2*a/(1+a'*D^-2*a)
    SigmaMult(a[k], ga, k,wgt);  // Sigma*a
    for(m = 0; m < mm; m++) {
      //ga[m] = 2*ga[m] - 2*a[k][m]/(diags[k][m]*diags[k][m]); // 2*Sigma*a - 2*D^-2*a
      //ga[m] += 2*a[k][m]*aDa/(diags[k][m]*diags[k][m]*denom); // + 2*D^-2*a*a'*D^-2*a/(1+a'*D^-2*a)
      ga[m] = 2*ga[m] - 2*tmp[m]/a[k][m]; // 2*Sigma*a - 2*D^-2*a
      ga[m] += 2*tmp[m]*aDa/(a[k][m]*denom); // + 2*D^-2*a*a'*D^-2*a/(1+a'*D^-2*a)
    }    

    // find diagonal's gradient
    // gd = 2*diag(Sigma)*diag(D) - 2*D^-1 + 2*D*(D^-2*a)^2/(1+a'*D^-2*a)
    double gd1, gd2, gd3;
    gd1 = gd2 = gd3 = 0;
    //cout << "sig\tD\tgd1\tgd2\tgd3" << endl;
    for(m = 0; m < mm; m++) {
      gd[m] = 2*sig[k][m][m]*diags[k][m] - 2/diags[k][m]; // 2*diag(Sigma)*diag(D) - 2*D^-1
      //gd[m] += (2*a[k][m]*a[k][m])/(diags[k][m]*SQR(diags[k][m])*denom);
      gd[m] += (2*tmp[m])/(diags[k][m]*denom);      
    }
  }

  /*
    Function: GetError
    
    Determines the error of potential values for the LRPDD of the covariance matrix. 
    These potential values are determined by the current values ('diags' and 'a') plus
    a step ('x') along a given search direction ('da' and 'dd'). This function is used
    in the conjugate gradients method for updating the LRPDD parameters. 

    Parameters:

    da - search direction for low-rank perturbation values (size dd)
    dd - search direction for diagonal matrix values (size dd)
    k - mixture we are working with
    x - step size along search direction
    wgt - prior probability for this mixture

    Preconditions:

    - input vectors are of proper size
    - covariance parameters have been computed/loaded
    - depends on conditions of <SigmaMult>

    Postconditions:

    - returns a double value indicating the log determinant of the difference between Sigma and the 
    potential LRPDD of Sigma. 

   */
  double GetError(vector<double> &da, vector<double> &dd, int k, double x, double wgt) {
    double p,b,ldet;
    double err;
    int m;
    vector<double> newa(mm,0);

    // find log(det(D^2+a*a')))
    p = 0;
    for(m = 0; m < mm; m++) p += log(SQR(diags[k][m]+x*dd[m]));    

    b = 0;
    for(m = 0; m < mm; m++) {
      b += exp(p - log(SQR(diags[k][m]+x*dd[m])) + log(SQR(a[k][m]+x*da[m])));
    }

    ldet = log(b+ exp(p));

    // find trace (Sigma*(D^2+a*a'))
    err = 0;
    for(m = 0; m < mm; m++) { newa[m] = a[k][m]+x*da[m]; }
    SigmaMult(newa,olda[k],k,wgt);
    for(m = 0; m < mm; m++) {
      err += sig[k][m][m]*SQR(diags[k][m]+x*dd[m])+olda[k][m]*(a[k][m]+x*da[m]);
    }
    /*
    if(x == 0) {
      cout << "olda: ";
      for(m = 0; m < mm; m++) cout << olda[k][m] << " ";
      cout << endl;
      cout << "ldet: " << ldet << endl;
      cout << "err: " << err << endl;
      cout << "total: " << err - ldet << endl;
      }*/

    err -= ldet;

    return err;
  }

  /*
    Function: PrintModel

    Parameters: 

    modelfilename - string with the name of the file to write the model to

    Preconditions: 

    - modelfilename is not null
    - model parameters have been computed/loaded

    Postcondition:

    - file has been created with the given name containing the 
    model parameters in a format outlined in <Model File Format>. 

  */
  void PrintModel(char* modelfilename) {
    assert(modelfilename != NULL);
    
    // print out parameters
    ofstream model_out(modelfilename);
    model_out.precision(30);
    model_out << kk << endl;
    for(int k = 0; k < (int)kk; k++) {
      model_out << "Component " << k << endl;
      // print priors    
      model_out << "Prior: " << (double)frac[k] << endl;
      // print means
      model_out << "Mean: " << endl;   
      for(int m = 0; m < (int)mm; m++) 
	model_out << (double)means[k][m] << " ";
      model_out << endl;
    // print sigs
      model_out << "Sig: " << endl;
      for(int m = 0; m < (int)mm; m++) {
	for(int mp = 0; mp < (int)mm; mp++) {	
	  model_out << (double)sig[k][m][mp] << " ";
	}
	model_out << endl;
      }
      model_out << endl;
    }
    
    model_out.close();  
  }

  /*
    Function: PrintClusterAssignments

    Prints the cluster assignments for each data sample to a file with the given name. 

    Parameters: 
    
    clusterfile - string containing the name of the file to write the cluster assignments to.

    Preconditions: 
    
    - clusterfile is not null
    - posterior probs have been computed for all data samples and clusters
    - depends on conditions of <AssignClusters>

    Postcondition:

    - cluster assignments have been written to a file with the given name. The file is 
    of the format described in <Cluster Assignments File Format>.

  */
  void PrintClusterAssignments(char* clusterfile) {
    assert(clusterfile != NULL);

    // output clustering    
    vector<int> assigned_clusters;
    double max_p;
    int max_k;
    ofstream clusters(clusterfile);
    double alpha = 0.7;

    cout << "writing clusters" << endl;
    assert((int)nn <= (int)resp.size());
    for(int i = 0; i < (int)nn; i++) {        
      AssignClusters(resp[i],assigned_clusters,alpha);

      for(int k = 0; k < (int)assigned_clusters.size(); k++) {
	clusters << assigned_clusters[k] << " ";
      }
      clusters << endl;
    }
    clusters.close(); 

  }

  /*
    Function: AssignClusters

    Given a list of posterior probabilities for a data sample(probs) and a threshold value (alpha), 
    a list of assigned clusters is returned (assigned_cluster). 

    Parameters:

    probs - the posterior probabilities of this data sample belonging to each cluster (size kk)
    assigned_cluster -  a vector to hold the clusters the data sample is assigned (size > 0)
    alpha - a double in [0..1] that thresholds the clusters a data sample is assigned to. 
    If the posterior probability is greater than alpha * max. posterior probability, 
    the data sample is assigned to that cluster. So, alpha=0 assigns to all clusters 
    and alpha=1 assigns to only the cluster with the max. posterior probability. 

    Preconditions:

    - input vectors are of proper size    

    Postconditions:

    - assigned_cluster is of a size greater than 0. 
  */
  void AssignClusters(vector<double> probs, vector<int> &assigned_cluster, double alpha) 
    {
      vector<double> sorted_probs(probs.size(),0.0);
      assigned_cluster.clear();
  
      // sort probs for this sample
      copy(probs.begin(),probs.end(),sorted_probs.begin());
      sort(sorted_probs.begin(),sorted_probs.end()); 
      reverse(sorted_probs.begin(),sorted_probs.end());

      assert( (int)sorted_probs.size() > 0 );

      for(int i = 0; i < (int)sorted_probs.size(); i++) {
	if(sorted_probs[i] >= alpha*sorted_probs[0]) {
	  for(int j = 0; j < (int)probs.size(); j++) {
	    if( find(assigned_cluster.begin(),assigned_cluster.end(),j) != assigned_cluster.end()) 
	      continue;
	    if(sorted_probs[i] == probs[j]) assigned_cluster.push_back(j);
	  }      
	} else { break; }
      }
    
    }


  /*
    Function: PrintProbs

    Parameters: 

    probfilename - string containing the name of the file to write the data sample posterior probabilities to. 

    Preconditions: 

    - probfilename is not null
    - posterior probs have been computed for all data samples

    Postcondition:

    - a file with the given name has been created and contains the posterior probs in 
    a format described in <Prob File Format> 

  */

  void PrintProbs(char* probfilename) {
    assert(probfilename != NULL); 

    cout << "writing probs" << endl;
    ofstream probfile(probfilename);
    probfile.precision(30);    
    for(int i = 0; i < (int)nn; i++) {    
      for(int k = 0; k < kk; k++) {
	probfile << resp[i][k] << " ";
      }
      probfile << endl;
    }

    probfile.close();        
  }
  
  /*
    Function: PrintSortedProbs

    A file is created that contains all possibly data/mixture pairs, in order of 
    thier relative posterior probability. The relative posterior probability is 
    the probability of a data sample, given a mixture, divided by the data 
    sample's maximum posterior probability over all mixtures. In other words, 
    a data sample's posterior probabilites are divided by the maximum norm of 
    those probabilities. This file is intended to be used with another 
    program in order to cluster the data into clusters with higher metric quality
    than those found by <AssignClusters>.

    Parameters: 

    probfilename - string containing the name of the file to write the data sample posterior probabilities to. 
    namefilename - string containing the name associated with each data sample.

    Preconditions: 

    - probfilename, namefilename is not null
    - posterior probs have been computed for all data samples

    Postcondition:

    - a file with the given name has been created and contains the posterior probs in 
    a format described in <Sorted Prob File Format> 

  */

  void PrintSortedProbs(char* probfilename, char* namefilename) {
    assert(probfilename != NULL);
    assert(namefilename != NULL);

    // remove data, since it isn't needed at this point
    cout << "clearing data to make room for sorting... ";
    data.clear();
    cout << "cleared." << endl;

    ifstream namefile(namefilename);
    ofstream probfile(probfilename);
    vector<string> names; 
    multimap<double,string> probs;
    stringstream outstr;
    int count = 0, countp= 0;

    cout << "sorting relative posterior probs" << endl;

    // read in names
    int num_data, num_neighbors;
    string name;
    namefile >> num_data; 
    while(!namefile.eof()) {
      namefile >> name >> num_neighbors; 
      names.push_back(name);
      countp++;    
    }
    namefile.close();
//   cout << "contents of names" << endl ;
  // for (int iter=0; iter < countp ; iter++) {
//	cout << names[iter] <<" ";
//   }
//   cout << endl << nn << endl ;

    cout <<" process probs" << endl ;
    double max_prob, prob_val;         
    count = 0 ;
  
    for(int n = 0; n < nn; n++) {
      // find maximum posterior prob
      max_prob = -1;
      for(int k = 0; k < kk; k++) {
	if(resp[n][k] > max_prob) max_prob = resp[n][k];
      }
      
//     cout << "calculate ratios to max prob, add to list of probs" << endl ;
      for(int k = 0; k < kk; k++) {
	 outstr.str("") ;
	 outstr << names[count] << " " << k;
	 prob_val = resp[n][k]/max_prob ;
	 string str2 = outstr.str();    
	probs.insert(pair<double,string>(prob_val,str2));
//	cout << " problem here ?" << endl ;
      }    

      // keep track of which data point we're working on
      count++;
    }

    // output sorted posterior probs 
    cout << "writing sorted probs" << endl;
    multimap<double,string>::reverse_iterator it;
    for(it = probs.rbegin(); it != probs.rend(); it++) {
      probfile << fixed << it->first << " " << it->second << endl;
    }
    probfile.close();

  }
};

/* 
   Function: LRP_CG

   This is wrapper function that allows Gauss_model to use ALGLIB functions. 

   Parameters:

   x - ALGLIB format of the covariance matrix parameters
   func - the output parameter for the error function 
   grad - ALGLIB format output parameter to hold the gradient with parameters defined by x
   ptr - holds a point to Gauss_model that contains the model being trained

   
 */
void LRP_CG(const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr)
{
  vector<double> ga,gd;
  Gauss_model *gmm = (Gauss_model*)ptr;

  int k = gmm->myk;
  double wgt = gmm->mywgt;

  // copy values to vars
  for(int m = 0; m < gmm->mm; m++) {
    gmm->a[k][m] = x[m];
    gmm->diags[k][m] = x[m+gmm->mm];
  }

  // calc function value
  func = gmm->GetError(gmm->a[k],gmm->diags[k], k, 0.0, wgt);

  // calc gradient
  gmm->GetGradient(ga,gd,k,wgt);

  // copy gradient info into alglib type
  for(int m = 0; m < gmm->mm; m++) {
    grad[m] = ga[m];
    grad[m+gmm->mm] = gd[m];
  }
}


#endif
