/*
Copyright (c) 2010 Jonathan Purnell 

This file is part of Cluster.

Cluster is free software and distributed under the terms of the 
GNU General Public License; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Cluster is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Cluster, named COPYING.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
  File: Cluster.cpp

  This is the library file for the Cluster program. 

  The Cluster program is designed to be exectued after a graph has been 
  embedded (using SSDE) and then used to train a Gaussian Mixture Model (GMM). 
  Given the graph of nodes and their corresponding posterior probabilities with
  respect to the mixtures from the Gaussian Mixture Model (GMM), this program 
  clusters the nodes in such a way that allows the cluster memberships to 
  overlap. Assignment of the nodes to clusters is based upon a cluster-based
  metric. The metric favors dense clusters (low average distance between nodes within the
  cluster) and distinct clusters (high average distance between nodes in different clusters).
  Here is an outline of the program
  
  - call readNodes()
   + create Node array to hold data and neighbors
   + calculated average degree of nodes (for GainThreshold)
  - call readFile()
   + store edge weights 
   + calculated average edge weight (for GainThreshold)
  - assign all nodes to most likely cluster (these are the first nn probs from input)
  - initialize Jaccards for each cluster
   + edges within cluster are added to to interal Jaccard
   + edges between clusters are added to each cluster's external Jaccard
  - process remaining probs
   + determine expected gain of adding a node to cluster (GainThreshold)
   + if gain of adding paper is greater than expected gain, add to cluster, otherwise ignore (AddPaper)
  - call PrintAssignments()
  - call PrintJaccards()
  And here is an outline of the other called functions:
  - GainThreshold()
   + Calculates the expected gain in adding a node to the cluster
  - AddPaper()
   + store current jaccards
   + add paper to cluster
   + if gain is higher than expected gain, return true
   + if gain is lower than expected gain, restore values and return false
*/


// Spectral Graph Embedding Method
// Eli Bocek-Rivele
// boceke@cs.rpi.edu
// 4/26/05

#include <cstdio>
#include <cstdlib>
#include <csignal>
#include <assert.h>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include <deque>
#include <map>
#include <time.h>
#include <vector>

#ifndef CLUSTER_H
#define CLUSTER_H

#define INFINITY_SSDE 100000000
#define SMALL 0.000000001


//#define SAMPLE_SIZE 25
//#define SVD_TOL 0.000000000001	// 10^-9 at least	
//#define PI_TOL 0.0000001	// Power iteration tolerance

#define RANDOM_SAMPLING 0
#define GREEDY_SAMPLING 1

// #define NUM_OF_SVS 25		// how many singular values are to be used(inverted) related with INV_TOL also
//#define SV_PERCENTAGE 0.9	// same concept as the one above

int NUM_DIMS = 5;
int SAMPLE_SIZE = 25;
double SVD_TOL = 0.000000000001; // 1e-12
double PI_TOL = 0.0000001; // 1e-7
double REG_PARAM = 3;		// this is the power of the first singular values in regularization parameter
int SAMPLING = GREEDY_SAMPLING;
int NUM_CLUSTERS = 10;

/* IMPORTANT NOTE: We are getting negative eigenvalues when SV_PERCENTAGE is low! 
 There must be a clever way of determining this percentage.
 This has to do with how many inverted singular values we have to collect in the pseudo-inverse of PHI 
 If we just choose a couple of top inverted singular values, then the product C*PHI_INVERSE*R may not
 be positive definite, leading to negative eigenvalues in the power iteration. On the other hand, 
 we cannot take all the inverted singular values due to numerical instability of the SVD procedure */

/* The number of inverted singular values we have to take is directly related to the numerical rank of the distance
matrix associated with the graph. If we take less than or more than that, the drawing is not pleasing */

using namespace std;

struct Node {
  //int index;			// the number assigned to this vertex
  string name;
  int Tindex;
  int lowlink;
  bool stacked; 

  int *neighbors;	// array to keep the indices of neighbors
  double *distances;       // array to keep the distances of neighbors
  int neighborSize;	// the index to keep the current size of the neighbor array
  int maximumSize;
  vector<int> member;

  // for BFS
  double distance;	// distance to the root

  Node(string name) {
    this->name = name;
    this->neighbors = new int[5];
    this->distances = new double[5];
    this->neighborSize = 0;
    this->maximumSize = 5;
    this->distance = (double)INFINITY_SSDE;
    this->Tindex = INFINITY_SSDE;
    this->stacked = false;
  }

  Node(string name, int num_neighbors) {
    this->name = name;
    this->neighbors = new int[num_neighbors];
    this->distances = new double[num_neighbors];
    this->neighborSize = 0;
    this->maximumSize = num_neighbors;
    this->distance = (double)INFINITY_SSDE;
    this->Tindex = INFINITY_SSDE;
    this->stacked = false;
  }

  ~Node() { 
    delete neighbors;
    delete distances;
  }
  
  bool IsMember(int cluster) {
    int i;
    for(i = 0; i < this->member.size(); i++) {
      if(this->member[i] == cluster) 
	return true;
    }
    return false;
  }
	
  void insert(int neighbor, double dist) {
    if(this->neighborSize == this->maximumSize)
      this->resize(2*this->maximumSize);

    neighbors[neighborSize] = neighbor;
    distances[neighborSize++] = dist;
  }

  void resize(int newSize) {
    int *oldArray = neighbors;
    double *oldDist = distances;
    neighbors = new int[newSize];
    distances = new double[newSize];

    for(int i = 0; i < maximumSize; i++ ) {
      neighbors[i] = oldArray[i];
      distances[i] = oldDist[i];
    }

    delete[] oldArray;
    delete[] oldDist;
    maximumSize = newSize;
  }
};

FILE *f = NULL;
FILE *g = NULL;
//double *xCoords, *yCoords;
vector< double* > Coords;
int numberOfNodes = 0;
Node **nodes = NULL;				// node array

int *samples = NULL;				// sampled vertices
double *minDistances = NULL;		
double **C = NULL;						// column sampling matrix
double **PHI = NULL;				// intersection of column and row sampling matrices
double **INV = NULL;				// inverse of PHI


double avgDegree;
double avgWeight;
int smallestDegreeNode;				// index of the smallest degree node
int firstSample;		// the first sampled node

map<string, int> node_map; // maps the node names to their index in nodes[]

double dotProduct(double a[], double b[]){
	double product = 0.0;
	for(signed int i=1; i <= numberOfNodes; i++){
		product += a[i]*b[i];
	}
	return product;
}

double dotProduct(vector<double> &a, vector<double> &b) {
	double product = 0.0;
	for(signed int i=1; i <= numberOfNodes; i++){
		product += a[i]*b[i];
	}
	return product;
}

double mag(double a[]) {
	return sqrt(dotProduct(a, a));
}

double mag(vector<double> &a) {
	return sqrt(dotProduct(a, a));
}

double* normalize(double a[]) {

	double magnitude = mag(a);

	for(signed int i=1; i <= numberOfNodes; i++){
		a[i] = a[i]/magnitude;
	}

	return a;
}

void normalize(vector<double> &a) {

	double magnitude = mag(a);

	for(signed int i=1; i <= numberOfNodes; i++){
		a[i] = a[i]/magnitude;
	}

	//return a;
}

double* multCons(double a[], double constant) {
	double *b = new double[numberOfNodes+1];
	for(signed int i=1; i <= numberOfNodes; i++){
		b[i] = a[i]*constant;
	}

	return b;
}

double* minusSSDE(double a[], double b[]) {
	double *c = new double[numberOfNodes+1];

	for(signed int i=1; i <= numberOfNodes; i++){
		c[i] = a[i]-b[i];
	}

	return c;
}

double* multGamma(double a[]) {
	double *b = new double[numberOfNodes+1];
	double sum = 0.0;

	for(signed int i=1; i <= numberOfNodes; i++){
		sum += a[i];
	}

	sum *= -1.0/(double)numberOfNodes;

	for(signed int i=1; i <= numberOfNodes; i++){
		b[i] = sum + a[i];
	}

	return b;
}

  //V2 = minusSSDE(V2, multCons(V1, (dotProduct(V1, V2))));
void makeOrtho(double a[], double b[]) { 
  double* c = multCons(b, dotProduct(b,a));
  
  for(signed int i=1; i <= numberOfNodes; i++) { a[i] -= c[i]; }

  delete [] c;
}

// multiply by C*INV*R
double* multL(double a[]) {
	double *b = new double[SAMPLE_SIZE+1];	// result of first multiplication
	double *c = new double[SAMPLE_SIZE+1];	// result of second multiplication

	double sum = 0.0;

	// multiply by C^T
	for(signed int i = 1; i <= SAMPLE_SIZE; i++) {
		sum = 0.0;
		for(int j = 1; j <= numberOfNodes; j++) 
			sum += -0.5*C[j][i]*a[j];
		b[i] = sum;
	}

	// multiply the result by INV
	for(int i = 1; i <= SAMPLE_SIZE; i++) {
		sum = 0.0;
		for(int j =	1; j <= SAMPLE_SIZE; j++) 
			sum += INV[i][j]*b[j];
		c[i] = sum;
	}

	// multiply the result by C
	for(int i = 1; i <= numberOfNodes; i++) {
		sum = 0.0;
		for(int j =	1; j <= SAMPLE_SIZE; j++) 
			sum += C[i][j]*c[j];
		a[i] = sum;
	}

	// free memory
	delete [] b;
	delete [] c;

	return a;
}


double* multGammaLGamma(double a[]) {
  double *b,*c;

  b = multGamma(a);
  b = multL(b);
  c = multGamma(b);
  
  delete [] b;
  
  return c;
}

/*
  Function: readNodes

  Loads the list of nodes names.  The format of this file is outlined in <Node Name File>.
  A hash table  is create so that the node names can be
  translated to their index in nodes[] in constant time.

*/
void readNodes(char *fileName) {
  ifstream fh(fileName);
  string nodeName;
  int num_neighbors;
  avgDegree = 0;
  avgWeight = 0;

  fh >> numberOfNodes;
  
  // prepare space for nodes
  nodes = new Node*[numberOfNodes+1];
  //xCoords = new double[numberOfNodes+1];
  //yCoords = new double[numberOfNodes+1];
  Coords.resize(NUM_DIMS,NULL);
  for(int dim = 0; dim < NUM_DIMS; dim++) {
    Coords[dim] = new double[numberOfNodes+1];
  }

  srand(time(0));
  for(int i = 1; i <= numberOfNodes; i++) {
    if( (i % 100000) == 0 ) { cout << (double)i/numberOfNodes << "%\r"; } 
    fh >> nodeName >> num_neighbors;
    nodes[i] = new Node(nodeName,num_neighbors);
    //xCoords[i] = rand();
    //yCoords[i] = rand();
    for(int dim = 0; dim < NUM_DIMS; dim++) {
      Coords[dim][i] = rand();
    }
    node_map[nodeName] = i;
    avgDegree += num_neighbors;
  }
  cout << endl;

  avgDegree /= numberOfNodes;
  cout << "Read " << numberOfNodes << " nodes" << endl;
  cout << "Avg Degree = " << avgDegree << endl;
}

/*
  Function: readFile 
  
  Loads the graph data. Format of this file follows <Graph File>.
*/
void readFile(char *fileName) {
  int z, numNodes, count;
  string x,y;
  double w;
  //f = fopen(fileName, "r");
  ifstream edgeFile(fileName);

  //fscanf(f, "%d", &numNodes);
  /*
  edgeFile >> numNodes;
  if(numberOfNodes != numNodes) {
    cout << "Wrong number of nodes" << endl;
    exit(0);
  }
  */
	
  //while(fscanf(f, "%s %s %f", &x, &y, &w) != EOF) {
  count = 0;
  while( edgeFile >> x >> y >> w ) {

    // Insert them into the array also (will be used in average and scale)	  	 
    count++;

    if( node_map.find(x) == node_map.end() ) {
      cout << "Node " << x << " not found for edge with " << y << endl;
      continue;
    }
    if( node_map.find(y) == node_map.end() ) {
      cout << "Node " << y << " not found for edge with " << x << endl;
      continue;
    }
    nodes[ node_map[x] ]->insert( node_map[y],1./w);
    nodes[ node_map[y] ]->insert( node_map[x],1./w);
    avgWeight += 1./w;
  }
  cout << endl;
  //fclose(f);
  edgeFile.close();
  avgWeight /= count;
}

/*
  Function: AddPaper

  The given paper is added to a given cluster. The gain of this addition is compared to the
  given expected gain of adding a paper to this cluster. If the addition of this paper yields
  a higher gain than the expected gain, the paper is added, the cluster metric updated and 
  a true value returned. Otherwise, the effects of adding the paper are undone and a false
  value is returned. 

 */
bool AddPaper(double prob, string paper_id, int cluster, vector<double> &Jint, vector<double> &Jext,
	      vector<int> &mem_count, double rand_gain, double &gain, double lambda) {

  double old_avgJac, new_avgJac;
  double old_ratio, new_ratio;
  double old_Jint, old_Jext;
  double this_gain;
  int ind;

  if(NUM_CLUSTERS != Jint.size()) { NUM_CLUSTERS = Jint.size();}

  // get index for node
  ind = node_map[paper_id];
  // calc current edge ratio
  old_ratio = Jint[cluster]/(Jint[cluster]+Jext[cluster]);
  // calc current cluster density
  old_avgJac = Jint[cluster]/(mem_count[cluster]*(mem_count[cluster]-1)); 
  // save old values
  old_Jint = Jint[cluster]; old_Jext = Jext[cluster];

  // update values based on adding paper
  // update member count of cluster
  mem_count[cluster]++;
  // adjust edge averages
  for(int j = 0; j < nodes[ ind ]->neighborSize; j++) {
    if(nodes[ nodes[ind]->neighbors[j] ]->IsMember(cluster)) {
      Jint[cluster] += nodes[ind]->distances[j];
      Jext[cluster] -= nodes[ind]->distances[j];
    } else {
      Jext[cluster] += nodes[ind]->distances[j];
    }
  }
  
  // calc new edge ratio
  new_ratio = Jint[cluster]/(Jint[cluster]+Jext[cluster]);
  // calc new cluster density
  new_avgJac = Jint[cluster]/(mem_count[cluster]*(mem_count[cluster]-1)); 
  // calc change in gain and density gain
  this_gain = (1-lambda)*(new_ratio - old_ratio) + lambda*(new_avgJac - old_avgJac);
  
  if(this_gain > rand_gain ) { // then add paper
    gain = this_gain;
    // add paper to cluster
    nodes[ind]->member.push_back(cluster);
    return true;
  } else { // reset values
    Jint[cluster] = old_Jint;
    Jext[cluster] = old_Jext;
    mem_count[cluster]--;
    return false;
  }  

  gain = this_gain;
  // add paper to cluster
  nodes[ind]->member.push_back(cluster);

  return true;
}

/*
  Function: GainThreshold

  Returns the expected gain of adding a paper to a cluster with the 
  given info and parameter value. 

  Parameters:

  Win - average internal edge weight of the cluster
  Wout - average external edge weight of the cluster
  c - the number of members in the cluster
  lambda - a float in [0,1] that adjusts the metric calculation. 
 */
double GainThreshold(double Win, double Wout, int c, double lambda) {
  double f = (double)c/(numberOfNodes - 1);
  double t1,t2,t3,t4;
  t1 = Win/(Win+Wout);
  t2 = (Win + avgWeight*avgDegree*f)/(Win+Wout+avgWeight*avgDegree*(1-f));
  t3 = Win/(c*(c-1)); 
  t4 = (Win+avgWeight*avgDegree*f)/(c*(c+1));
  
  return (1-lambda)*(t1-t2)+2*lambda*(t3-t4);
}

/*
  Function: PrintAssignments

  Prints the assignments of each node. 
 */
void PrintAssignments() {
  ofstream ofile("cluster_assign_randThreshold.txt");

  ofile << numberOfNodes << endl;
  
  for(int i = 1; i <= numberOfNodes; i++) {
    ofile << nodes[i]->name << " ";
    for(int j = 0; j < nodes[i]->member.size(); j++) {
      ofile << nodes[i]->member[j] << " ";
    }
    ofile << endl;
  }

  ofile.close();
}

/*
  Function: PrintJaccards

  Prints the Jaccards of internal and external weights
  for each pair of clusters. 
 */
void PrintJaccards() {
  // calculate Jaccards for each cluster internal and external
  vector<double> AA(NUM_CLUSTERS,0.0);
  double AC,AB,BC;
  int AAc,ACc,ABc,BCc;

  // setup internal values
  for(int cluster = 0; cluster < NUM_CLUSTERS; cluster++) {
    AAc = 0;
    for(int i = 1; i <= numberOfNodes; i++) {
      if(nodes[i]->IsMember(cluster)) {
	for(int j = 0; j < nodes[i]->neighborSize; j++) {
	  if(nodes[ nodes[i]->neighbors[j] ]->IsMember(cluster)) {
	    AA[cluster] += nodes[i]->distances[j];
	    AAc++;
	  }
	}
      }
    }
    AA[cluster] /= AAc;
    cout << cluster << " : " << AA[cluster] << endl;
  }
  
  for(int cluster = 0; cluster < NUM_CLUSTERS; cluster++) {
    for(int clusterB = cluster+1; clusterB < NUM_CLUSTERS; clusterB++) { // process all cross overs	
      AC=AB=BC=0;
      ACc=ABc=BCc=0;

      for(int i = 1; i <= numberOfNodes; i++) { // search all nodes
   
	if( nodes[i]->IsMember(cluster) && !nodes[i]->IsMember(clusterB) ) { // a node in A
	  for(int j = 0; j < nodes[i]->neighborSize; j++) {
	    // if neighbor is internal
	    if(nodes[ nodes[i]->neighbors[j] ]->IsMember(clusterB)) { 
	      if( nodes[nodes[i]->neighbors[j]]->IsMember(cluster)) { // a node in C
		ACc++;
		AC += nodes[i]->distances[j];
	      } else { // a node in B
		ABc++;
		AB += nodes[i]->distances[j];
	      }
	    } 
	  }
	}

	if( !nodes[i]->IsMember(cluster) && nodes[i]->IsMember(clusterB) ) { // a node in B
	  for(int j = 0; j < nodes[i]->neighborSize; j++) {
	    // if neighbor is internal
	    if(nodes[ nodes[i]->neighbors[j] ]->IsMember(cluster)) { 
	      if( nodes[nodes[i]->neighbors[j]]->IsMember(cluster)) { // a node in C
		BCc++;
		BC += nodes[i]->distances[j];
	      } else { // a node in A
		ABc++;
		AB += nodes[i]->distances[j];
	      }
	    } 
	  }
	}
      }
      AC /= ACc; AB /= ABc; BC /= BCc;
      cout << "(" << cluster << "," << clusterB << "): ";
      cout << 0.5*(AA[cluster]+AA[clusterB]) << " " << 0.5*(AC+BC) << " " << AB << endl;
    }    
  
  }
}


/* 
   Function: AverageAssignmentClustering

   Data samples are assigned to their most likely clusters based on the given 
   posterior probabilities. Assignments are conducted in order of decreaseing 
   relative posterior probability (see documentation on the GMM code) and 
   continue until the given average of assigned clusters (lambda) is achieved. 

   Parameters:

   lambda - indicates the average number of clusters a data point is assigned to. 

   Precondition:

   - lambda must be greater than or equal to 1

   Postcondition:

   - The average number of clusters a data sample is assigned to is as close to 
   lambda as possible for the given number of data samples and clusters. 
   - if lambda exceeds the maximum possible average (i.e. the number of clusters),
   then all data are assigned to all clusters. 
 
 */
void AverageAssignmentClustering(double lambda) {

  // average assignment
  double total_assign;
  vector<int> mem_count;
  double prob;
  string paper_id;
  int cluster;

  ifstream probfile;
  probfile.exceptions(ifstream::failbit);
  try{ probfile.open("sorted_probs.txt"); }
  catch (ifstream::failure e) { 
    cout << "ifstream error: sorted_probs.txt probably doesn't exist" << endl;
    exit(0);
  }

  // total number of assignments should be between numberOfNodes (each data sample
  // assigned to most likely cluster) and numberOfNodes*NUM_CLUSTERS (every data sample
  // is assigned to every cluster). 
  assert(lambda >= 1.0);
  if(lambda > NUM_CLUSTERS) lambda = NUM_CLUSTERS;
  total_assign = int(numberOfNodes * lambda); 
    
  for(int i = 0; i < total_assign; i++) {
    probfile >> prob >> paper_id >> cluster;
    nodes[ node_map[paper_id] ]->member.push_back(cluster);
    if(mem_count.size() < cluster) { mem_count.resize(cluster); }
    mem_count[cluster]++;
  }
    
  probfile.close();
  
}

/*
  Function: MetricBasedClustering

  Data samples are processed in order of decreasing relative posterior probability
  (see documentation on GMM for futher details). For each possible assignment, the
  expected improvement of assigning a data sample to the cluster is calculated. A
  data sample is assigned only if the actual gain in assigning it to the cluster 
  is greater than the expected improvement. Thus, we ensure that, relative to the
  metric, the clustering is as good or better than partitioning the data samples 
  into their most likely clusters. 

  Parameters:

  lambda - a parameter for the metric (see <GainThreshold> or <AddPaper>)
  
  Precondition:

  - the file 'sorted_probs.txt' exists

  Postcondition:

  - the data samples are assigned in such a way to improve the clusters
  based on a metric. 
 */
void MetricBasedClustering(double lambda) {


  vector<double> Jint, Jext;
  vector<int> mem_count; 
  double prob;
  string paper_id;
  int cluster;
  ofstream resfile("alphaprob.txt");

  double gain = 0;
  double Jacgain = 0;
  double rand_gain;
  int counter = 0;

  ifstream probfile;
  probfile.exceptions(ifstream::failbit);
  try{ probfile.open("sorted_probs.txt"); }
  catch (ifstream::failure e) { 
    cout << "ifstream error: sorted_probs.txt probably doesn't exist" << endl;
    exit(0);
  }

  // get initial memberships
  while(!probfile.eof()) {
    try{
      probfile >> prob;
    } catch (ios_base::failure& e) {
      break;
    }
    
    try{
      probfile >> paper_id >> cluster;
    } catch (ios_base::failure& e) {
      cerr << "Something really went wrong reading the sorted_probs file" << endl;
      exit(11);
    }
    
    if(prob < 1.0) { break; }
    nodes[ node_map[paper_id] ]->member.push_back(cluster);
    if(mem_count.size() < cluster + 1) { mem_count.resize(cluster + 1); }
    mem_count[cluster]++;
  }

  // resize cluster params
  Jint.resize(mem_count.size());
  Jext.resize(mem_count.size());

  // initialize Jaccards
  for(int i = 1; i <= numberOfNodes; i++) {
    // go through neighbors
    for(int j = 0; j < nodes[i]->neighborSize; j++) {
      // if neighbor is internal
      if(nodes[ nodes[i]->neighbors[j] ]->member[0] == nodes[i]->member[0]) {
	Jint[ nodes[i]->member[0] ] += 0.5*nodes[i]->distances[j];
      } else { // if neighbor is external
	Jext[ nodes[i]->member[0] ] += nodes[i]->distances[j];
      }
    }
  }
  
  // process read in prob  
  rand_gain = GainThreshold(Jint[cluster], Jext[cluster], mem_count[cluster], lambda);
  if(AddPaper(prob,paper_id,cluster,Jint,Jext,mem_count,rand_gain,gain,lambda)) {
    resfile << paper_id << " " << prob << " " << gain << " " << rand_gain << endl;
  } else {
    resfile << "ignored " << paper_id << endl;
  }
  //AddPaper(prob,paper_id,cluster,Jint,Jext,mem_count,gain,lambda);
  //resfile << paper_id << " " << prob << " " << gain << endl;


  // process remaining probs
  cout << "process remaining probs" << endl;
  while(!probfile.eof()) {
    probfile >> prob >> paper_id >> cluster;     

    if( prob <= 0.0 ) { break; }

    rand_gain = GainThreshold(Jint[cluster], Jext[cluster], mem_count[cluster], lambda);
    if(AddPaper(prob,paper_id,cluster,Jint,Jext,mem_count,rand_gain,gain,lambda)) {
      resfile << paper_id << " " << prob << " " << gain << " " << rand_gain << endl;
    } else {
      resfile << "ignored " << paper_id << " for cluster " << cluster << endl;
    }
    //AddPaper(prob,paper_id,cluster,Jint,Jext,mem_count,gain,lambda);
    //  resfile << paper_id << " " << prob << " " << gain << endl;   
  }

  resfile.close();  
  probfile.close();
  
  PrintJaccards();
}

/*
  Function: PostProcessPruning

  Removes data samples from assigned clusters if the corresponding 
  posterior probability falls outside a user-defined threshold. The
  mean and variance of the posterior probabilities is calculated. 
  If a data sample has a posterior probability for an assigned cluster
  that is more than a given number of deviations from the mean posterior
  probability, the data sample is unassigned from that cluster. This
  can help catch outlier data samples that do not belong to any cluster
  in particular. 

  Parameters:
  
  prune - the multiple of deviations within which a data sample's post. prob. 
  must fall to remain assigned to a cluster. If the post prob is outside 
  the mean +/- prune*std.dev., the data sample is unassiged from the cluster. 

  Preconditions:

  - prune should be a positive value

  Postcondition:

  - based on the given pruning value, certain data samples may be unassiged
  from certain clusters. 
 */
void PostProcessPruning(double prune) {

  // calculate average/variance of posterior probs
  ifstream postprobs("probs.txt");
  double avgprob, varprob, newprob, delta, sqrsum;
  int count = 0;
  avgprob = varprob = sqrsum = 0;

  while(!postprobs.eof()) {
    postprobs >> newprob; 
    count++;
      
    delta = newprob - avgprob;
    avgprob += delta/(double)count; 
    sqrsum += delta*(newprob - avgprob);
    varprob  = sqrsum/(double)count;
  }

  // reset file pointer to beginning of file
  postprobs.seekg(0,ios::beg);

  // find which post probs are outside determine pruning limits
  vector<int>::iterator it;
  count = 1; 
  while(!postprobs.eof()) {
    for(int k = 0; k < NUM_CLUSTERS; k++) {
      postprobs >> newprob; 
      delta = newprob - avgprob; 
      if(delta < 0) delta *= -1; // make positive
	
      if( delta > (prune*varprob)) {  // if post prob is beyond pruning threshold
	it = find(nodes[count]->member.begin(), nodes[count]->member.end(), k);
	if( it != nodes[count]->member.end() ) 
	  nodes[count]->member.erase(it);	  
      }
    }
    count++;
  }

  postprobs.close();
}

#endif
