/*
Copyright (c) 2010 Jonathan Purnell 

This file is part of Cluster.

Cluster is free software and distributed under the terms of the 
GNU General Public License; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Cluster is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Cluster, named COPYING.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
  File: main.cpp
  
  This is the main file for the Cluster program. 
 */
#include <cstdio>
#include <cstdlib>
#include <csignal>
#include <assert.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include <deque>
#include <map>
#include <time.h>
#include <vector>
#include "Cluster.h"

using namespace std;

void AverageAssignmentClustering(double lambda);
void MetricBasedClustering(double lambda);

/*
  Function: main

  The main function for the Cluster program.

  Parameters:

  node names - a file that holds the name for each node/data sample
  graph file - a file that holds all the edges and their weights in the original graph
  num of clusters - number of clusters/mixtures in the model that generated the posterior probabilities
  cluster method - if 0, the metric-based clustering is performed. if 1, average assignment method is used
                   (see <Cluster.h> for details on methods)
  lambda - parameter for the clustering method. 
  pruning param - optional. if used, the clustering is post-processed by pruning. This parameter determines the 
                  extent of the pruning. 
*/
int main(int argv, char **argc){
  if(argv <= 2) {
    cout << "usage: Cluster <node names> <graph file> <num of clusters> <cluster method> <lambda> [pruning param]" << endl;
    cout << "cluster method: (0 - metric-based , 1 - avg. assignment)" << endl;
    cout << "lambda: " << endl;
    cout << "  metric-based clustering: metric parameter" << endl;
    cout << "  avg. assignment: average number of clusters a data sample is assigned to" << endl;
    cout << "pruning param: data sample/cluster pairs that have a posterior probability outside " << endl;
    cout << "               the given number of standard deviations from the mean posterior prob. " << endl;
    cout << "               are ruled out for possible clustering (see documentation for details)." << endl;
    exit(0);
  }
  
  cout << "Cluster ";
  for(int i =1; i < argv; i++) { cout << argc[i] << " "; }
  cout << endl;


  NUM_CLUSTERS = atoi(argc[3]);
  double lambda = atof(argc[5]);  
  
  cout << "Reading names" << endl;
  readNodes(argc[1]);
  cout << "Reading edges" << endl;
  readFile(argc[2]);
  cout << "Avg weight = " << avgWeight << endl;
  

  switch(atoi(argc[4])) { 
  case 0: // metric-based clustering
    MetricBasedClustering(lambda);
    break;
  case 1: 
    AverageAssignmentClustering(lambda); 
    break;
  }   

  if( argv >= 7 ) { // post pruning
    double prune = atof(argc[6]); // pruning param
    PostProcessPruning(prune);
  }

  PrintAssignments();

  // delete data 
  delete[] nodes;
  
  return 0;
}

