
#include <cstdio>
#include <cstdlib>
#include <csignal>
#include <assert.h>
#include <iostream>
#include <algorithm>
#include <fstream>
#include <cmath>
#include <deque>
#include <map>
#include <time.h>
#include <vector>
#include "SSDE.h"

/*
  Function: main

  The main function for the Sampled Spectral Distance Embedding program. 

  Parameters:

  - first parameter is the name of the executable
  - second parameter is the name of the graph
  - third parameter is the mode of execution (assign random coordinates or run)
  - fourth parameter is the SAMPLING method (whether random or clever)
  - fifth parameter signals to print or not print the C matrix and the inverse of the Phi matrix (0 - do not print, 1 - print).
    When this parameter is set to 1, the program will produce C.txt and InvPhi.txt. 
  - [optional] sixth parameter is DIMS (the # of dimensions to embed the graph into)
  - [optional] seventh parameter is PI_TOL (power iteration tolerance)
  - [optional] eighth parameter is SVD_TOL (tolerance used in SVD procedure)
  - [optional] nineth parameter is SAMPLE_SIZE
  - [optional] tenth parameter is REG_PARAM (power of the first singular value used in regularization of PHI inverse)
*/
int main(int argv, char **argc){
  if(argv < 6) {
    cout << "usage: SSDE <node names> <graph file> <exec. mode> <sampling> <print mats> [DIMS SAMPLE_SIZE PI_TOL SVD_TOL REG_PARAM] " << endl;
    cout << "exec. mode = rand|run [use run]" << endl;
    cout << "sampling = 0|1 (0=random,1=clever) [use 1]" << endl;
    cout << "print mats: (0 = do not print, 1 = print C and inverse of Phi) " << endl;
    cout << "DIMS = output dimensions" << endl;
    cout << "Ouput file: final_coords.txt" << endl;
    exit(0);
  }

  //system("date");

  for(int i = 0; i < argv; i++) {
    cout << argc[i] << " ";
  }
  cout << endl;

  bool printMats;
  printMats = ( atoi(argc[5]) == 1 );
  
  // read in parameters
  SAMPLING = atoi(argc[4]);
  if(argv >= 7) { 
    NUM_DIMS = atoi(argc[6]); 
    cout << "NUM_DIMS = " << NUM_DIMS << endl;
  } 

  cout << "Reading names" << endl;
  readNodes(argc[1]);
  cout << "Reading edges" << endl;
  readFile(argc[2]);
  
  if(argv >= 8) { SAMPLE_SIZE = atoi(argc[7]); }
  else {SAMPLE_SIZE = (int)min(numberOfNodes, max((int)log((double)numberOfNodes), 50));}
  if(argv >= 9) { PI_TOL = atof(argc[8]); }
  if(argv >= 10) { SVD_TOL = atof(argc[9]); }  
  if(argv >= 11) { REG_PARAM = atof(argc[10]); }

  if(strcmp(argc[3], "rand") == 0)
    writeFile("random_coords.txt");

  else {
    

    if( SAMPLING == 2 ) { // find largest component
      cout << "Finding largest component..." << endl;
      ofstream scc_file("sccs.txt");
      int index = 0;
      std::deque<int> S;
      for(int v = 1; v <= numberOfNodes; v++) {
	if( nodes[v]->Tindex == INFINITY_SSDE) {
	  Tarjan(v,index,S,scc_file);
	}
      }
      scc_file.close();
    } else { // run ssde
      // instantiate the matrices keeping all-pairs-shortest path
      // C is (n x c)
      C = new double*[numberOfNodes+1];
      for(int i = 1; i <= numberOfNodes; i++ )
	C[i] = new double[SAMPLE_SIZE+1];

      // perform BFS
      clock_t a = clock();
      allBFS(SAMPLING);
      clock_t b = clock();
      cout << "BFS: " << (b-a)/(double)CLOCKS_PER_SEC << " seconds" << endl;
      writeFile("BFSCoords.txt");
      
      // Sampling/SVD
      clock_t c = clock();
      sample();
      clock_t d = clock();
      cout << "SVD: " << (d-c)/(double)CLOCKS_PER_SEC << " seconds" << endl;
      writeFile("SVDCoords.txt");
      
      // output C and PHI
      if(printMats) PrintCandInvPhi();

      // Power iterations
      clock_t e = clock();
      powerIteration();
      clock_t f = clock();
      writeFile("PowerIterCoords.txt");
      cout << "Power Iteration: " << (f-e)/(double)CLOCKS_PER_SEC << " seconds" << endl;

      cout << "Total: " << (f-e+d-c+b-a)/(double)CLOCKS_PER_SEC << " seconds" << endl;

      normalizeCoords();
      cout << "writing coords" << endl;
      writeFile("final_coords.txt");
                 
      cout << "Deleting C" << endl;
      delete[] C;

      cout << "Deleting PHI and INV" << endl;
	  
	for(int i = SAMPLE_SIZE; i > 0; --i) 
		delete[] PHI[i];
	delete[] PHI;
	PHI = NULL;
	 
	  for(int i = SAMPLE_SIZE; i > 0; --i) {
		   delete[] INV[i];
	  }
      delete[] INV;
      
	}   
    
    cout << "Deleting Coords" << endl;
    for(int dim = 0; dim < NUM_DIMS; dim++) {
      delete[] Coords[dim];
    }

    cout << "Deleteing nodes" << endl;
    delete[] nodes;

    //system("~/personal/twitter_update.sh");      
  }

  return 0;
}
